"""Tests for ActiveCallsService."""
from __future__ import annotations

import pytest
from pytest_httpx import HTTPXMock

from tests.conftest import TOKEN_URL, TOKEN_JSON, api
from threecx import ThreeCXClient


CALL_1 = {"Id": 1, "Status": "Connected", "Established": True, "Caller": "100", "Callee": "200", "Duration": 30}
CALL_2 = {"Id": 2, "Status": "Ringing", "Established": False, "Caller": "101", "Callee": "0031612345678"}


@pytest.fixture
def client(httpx_mock: HTTPXMock) -> ThreeCXClient:
    httpx_mock.add_response(url=TOKEN_URL, method="POST", json=TOKEN_JSON, is_reusable=True)
    return ThreeCXClient("https://pbx.test", "id", "secret")


def test_list_active_calls(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(url=api("/ActiveCalls"), json={"value": [CALL_1, CALL_2]})
    calls = client.active_calls.list()
    assert len(calls) == 2
    assert calls[0].id == 1
    assert calls[0].duration == 30
    assert calls[1].caller == "101"


def test_list_active_calls_empty(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(url=api("/ActiveCalls"), json={"value": []})
    assert client.active_calls.list() == []


def test_iterate_active_calls(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(url=api("/ActiveCalls"), json={"value": [CALL_1, CALL_2]})
    calls = list(client.active_calls.iterate())
    assert len(calls) == 2


def test_drop_call(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=api("/ActiveCalls(1)/Pbx.DropCall"),
        method="POST",
        status_code=204,
    )
    client.active_calls.drop(1)


def test_active_call_on_hold_field(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=api("/ActiveCalls"),
        json={"value": [{**CALL_1, "OnHold": True, "Recording": False}]},
    )
    calls = client.active_calls.list()
    assert calls[0].on_hold is True
    assert calls[0].recording is False



from __future__ import annotations

from typing import Any, Dict, Iterator, List, Optional

from ..models.queues import Queue, QueueAgent, QueueManager
from ..odata import ODataQuery
from .base import BaseService


class QueuesService(BaseService):
    """CRUD operations for call queues.

    Reference paths:
      GET    /Queues
      POST   /Queues
      GET    /Queues({Id})
      GET    /Queues(Number={Number})
      PATCH  /Queues({Id})
      DELETE /Queues({Id})
      GET    /Queues({Id})/Agents
      GET    /Queues({Id})/Managers
      POST   /Queues({Id})/Pbx.ResetQueueStatistics
    """

    _PATH = "/Queues"

    def list(self, query: Optional[ODataQuery] = None) -> List[Queue]:
        data = self._list_raw(self._PATH, query)
        return [Queue.model_validate(item) for item in data.get("value", [])]

    def iterate(self, query: Optional[ODataQuery] = None) -> Iterator[Queue]:
        yield from self._paginate(self._PATH, Queue, query)

    def create(self, queue: Queue | Dict[str, Any]) -> Queue:
        payload = queue.model_dump(by_alias=True, exclude_none=True) if isinstance(queue, Queue) else queue
        data = self._post(self._PATH, json=payload)
        return Queue.model_validate(data)

    def get(self, queue_id: int, query: Optional[ODataQuery] = None) -> Queue:
        data = self._get(f"{self._PATH}({queue_id})", params=self._query_params(query))
        return Queue.model_validate(data)

    def get_by_number(self, number: str, query: Optional[ODataQuery] = None) -> Queue:
        data = self._get(f"{self._PATH}(Number='{number}')", params=self._query_params(query))
        return Queue.model_validate(data)

    def update(self, queue_id: int, changes: Queue | Dict[str, Any]) -> None:
        payload = changes.model_dump(by_alias=True, exclude_none=True) if isinstance(changes, Queue) else changes
        self._patch(f"{self._PATH}({queue_id})", json=payload)

    def delete(self, queue_id: int, etag: Optional[str] = None) -> None:
        self._delete(f"{self._PATH}({queue_id})", etag=etag)

    # ------------------------------------------------------------------
    # Sub-resources
    # ------------------------------------------------------------------

    def get_agents(self, queue_id: int) -> List[QueueAgent]:
        data = self._get(f"{self._PATH}({queue_id})/Agents")
        return [QueueAgent.model_validate(item) for item in data.get("value", [])]

    def get_managers(self, queue_id: int) -> List[QueueManager]:
        data = self._get(f"{self._PATH}({queue_id})/Managers")
        return [QueueManager.model_validate(item) for item in data.get("value", [])]

    def reset_statistics(self, queue_id: int) -> None:
        self._post(f"{self._PATH}({queue_id})/Pbx.ResetQueueStatistics")



from __future__ import annotations

from typing import Any, Dict, Iterator, List, Optional

from ..models.phones import Phone, PhoneTemplate
from ..odata import ODataQuery
from .base import BaseService


class PhonesService(BaseService):
    """Manage provisioned phones and phone templates.

    Reference paths:
      GET    /Phones
      GET    /Phones({Id})
      PATCH  /Phones({Id})
      DELETE /Phones({Id})
      GET    /PhoneTemplates
      GET    /PhoneTemplates({Id})
      POST   /Users/Pbx.RebootPhone
      POST   /Users/Pbx.ReprovisionPhone
    """

    _PHONES = "/Phones"
    _TEMPLATES = "/PhoneTemplates"

    # ------------------------------------------------------------------
    # Phones
    # ------------------------------------------------------------------

    def list(self, query: Optional[ODataQuery] = None) -> List[Phone]:
        data = self._list_raw(self._PHONES, query)
        return [Phone.model_validate(item) for item in data.get("value", [])]

    def iterate(self, query: Optional[ODataQuery] = None) -> Iterator[Phone]:
        yield from self._paginate(self._PHONES, Phone, query)

    def get(self, phone_id: int, query: Optional[ODataQuery] = None) -> Phone:
        data = self._get(f"{self._PHONES}({phone_id})", params=self._query_params(query))
        return Phone.model_validate(data)

    def update(self, phone_id: int, changes: Phone | Dict[str, Any]) -> None:
        payload = changes.model_dump(by_alias=True, exclude_none=True) if isinstance(changes, Phone) else changes
        self._patch(f"{self._PHONES}({phone_id})", json=payload)

    def delete(self, phone_id: int, etag: Optional[str] = None) -> None:
        self._delete(f"{self._PHONES}({phone_id})", etag=etag)

    # ------------------------------------------------------------------
    # Phone templates
    # ------------------------------------------------------------------

    def list_templates(self, query: Optional[ODataQuery] = None) -> List[PhoneTemplate]:
        data = self._list_raw(self._TEMPLATES, query)
        return [PhoneTemplate.model_validate(item) for item in data.get("value", [])]

    def get_template(self, template_id: int) -> PhoneTemplate:
        data = self._get(f"{self._TEMPLATES}({template_id})")
        return PhoneTemplate.model_validate(data)

    # ------------------------------------------------------------------
    # Phone actions (triggered via Users service path)
    # ------------------------------------------------------------------

    def reboot(self, mac: str) -> None:
        """Reboot a phone by MAC address."""
        self._post("/Users/Pbx.RebootPhone", json={"MAC": mac})

    def reprovision(self, mac: str) -> None:
        """Reprovision a phone by MAC address."""
        self._post("/Users/Pbx.ReprovisionPhone", json={"MAC": mac})



"""Tests for CallHistoryService, including multi-page iteration and count."""
from __future__ import annotations

import pytest
from pytest_httpx import HTTPXMock

from tests.conftest import TOKEN_URL, TOKEN_JSON, api
from threecx import ThreeCXClient
from threecx.odata import ODataQuery


ENTRY_1 = {"Id": "e1", "Caller": "100", "Callee": "200", "CallDirection": "Internal", "Status": "Answered", "TalkDuration": 60}
ENTRY_2 = {"Id": "e2", "Caller": "101", "Callee": "0031612345678", "CallDirection": "Outbound", "Status": "Answered", "TalkDuration": 120}
ENTRY_3 = {"Id": "e3", "Caller": "102", "Callee": "0031698765432", "CallDirection": "Outbound", "Status": "NoAnswer", "TalkDuration": 0}


@pytest.fixture
def client(httpx_mock: HTTPXMock) -> ThreeCXClient:
    httpx_mock.add_response(url=TOKEN_URL, method="POST", json=TOKEN_JSON, is_reusable=True)
    return ThreeCXClient("https://pbx.test", "id", "secret")


def test_list_call_history(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(url=api("/CallHistoryView"), json={"value": [ENTRY_1, ENTRY_2]})
    entries = client.call_history.list()
    assert len(entries) == 2
    assert entries[0].id == "e1"
    assert entries[1].direction == "Outbound"


def test_list_with_filter(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=api("/CallHistoryView"),
        match_params={"$filter": "CallDirection eq 'Outbound'"},
        json={"value": [ENTRY_2, ENTRY_3]},
    )
    entries = client.call_history.list(ODataQuery().filter("CallDirection eq 'Outbound'"))
    assert len(entries) == 2


def test_iterate_multiple_pages(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=api("/CallHistoryView"),
        json={"value": [ENTRY_1], "@odata.nextLink": api("/CallHistoryView?$skip=1")},
    )
    httpx_mock.add_response(
        url=api("/CallHistoryView?$skip=1"),
        json={"value": [ENTRY_2], "@odata.nextLink": api("/CallHistoryView?$skip=2")},
    )
    httpx_mock.add_response(
        url=api("/CallHistoryView?$skip=2"),
        json={"value": [ENTRY_3]},
    )
    entries = list(client.call_history.iterate())
    assert len(entries) == 3
    assert [e.id for e in entries] == ["e1", "e2", "e3"]


def test_count(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=api("/CallHistoryView"),
        match_params={"$count": "true", "$top": "0"},
        json={"@odata.count": 1337, "value": []},
    )
    assert client.call_history.count() == 1337


def test_count_with_filter(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=api("/CallHistoryView"),
        match_params={"$filter": "Status eq 'Answered'", "$count": "true", "$top": "0"},
        json={"@odata.count": 42, "value": []},
    )
    assert client.call_history.count(ODataQuery().filter("Status eq 'Answered'")) == 42



"""Tests for SystemService."""
from __future__ import annotations

import pytest
from pytest_httpx import HTTPXMock

from tests.conftest import TOKEN_URL, TOKEN_JSON, api
from threecx import ThreeCXClient


STATUS_PAYLOAD = {
    "HasActiveCalls": True,
    "CallsActive": 5,
    "ExtensionsRegistered": 20,
    "ExtensionsTotal": 25,
    "TrunksRegistered": 2,
    "TrunksTotal": 3,
    "CpuUsage": 15.2,
    "MemUsage": 42.0,
    "DiskUsage": 30.5,
    "Version": "20.0.0.123",
    "FQDN": "pbx.example.com",
}

LICENSE_PAYLOAD = {
    "ProductCode": "3CXPSPRO",
    "MaxSimCalls": 32,
    "MaxExtensions": 100,
    "RegisteredExtensions": 25,
}

PARAMS_PAYLOAD = {
    "FQDN": "pbx.example.com",
    "Name": "My PBX",
    "HttpPort": 5000,
    "HttpsPort": 5001,
    "SIPPort": 5060,
}


@pytest.fixture
def client(httpx_mock: HTTPXMock) -> ThreeCXClient:
    httpx_mock.add_response(url=TOKEN_URL, method="POST", json=TOKEN_JSON, is_reusable=True)
    return ThreeCXClient("https://pbx.test", "id", "secret")


def test_get_status(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(url=api("/SystemStatus"), json=STATUS_PAYLOAD)
    status = client.system.get_status()
    assert status.calls_active == 5
    assert status.pbx_version == "20.0.0.123"
    assert status.fqdn == "pbx.example.com"
    assert status.cpu_usage == pytest.approx(15.2)


def test_get_license(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(url=api("/LicenseStatus"), json=LICENSE_PAYLOAD)
    lic = client.system.get_license()
    assert lic.product_code == "3CXPSPRO"
    assert lic.max_simultaneous_calls == 32
    assert lic.max_extensions == 100


def test_get_parameters(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(url=api("/Parameters"), json=PARAMS_PAYLOAD)
    params = client.system.get_parameters()
    assert params.fqdn == "pbx.example.com"
    assert params.sip_port == 5060


def test_get_raw(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(url=api("/EventLogs"), json={"value": [{"Id": 1, "Message": "Started"}]})
    data = client.system.get_raw("/EventLogs")
    assert data["value"][0]["Message"] == "Started"


def test_post_raw(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=api("/Backups/Pbx.CreateBackup"),
        method="POST",
        status_code=204,
    )
    result = client.system.post_raw("/Backups/Pbx.CreateBackup")
    assert result is None



"""Unit tests for the exception hierarchy and raise_for_status."""
from __future__ import annotations

import pytest

from threecx.exceptions import (
    AuthenticationError,
    NotFoundError,
    RateLimitError,
    ServerError,
    ThreeCXError,
    ValidationError,
    _extract_message,
    raise_for_status,
)


# --- _extract_message ---------------------------------------------------------

def test_extract_message_plain_string():
    assert _extract_message("oops") == "oops"


def test_extract_message_dict_with_error_string():
    assert _extract_message({"error": "invalid_client"}) == "invalid_client"


def test_extract_message_dict_with_error_object():
    body = {"error": {"message": "User not found", "code": "404"}}
    assert _extract_message(body) == "User not found"


def test_extract_message_dict_no_error_key():
    body = {"someOtherKey": "value"}
    result = _extract_message(body)
    assert "someOtherKey" in result


def test_extract_message_non_string():
    assert _extract_message(42) == "42"


# --- raise_for_status ---------------------------------------------------------

@pytest.mark.parametrize("code,exc_type", [
    (401, AuthenticationError),
    (404, NotFoundError),
    (400, ValidationError),
    (422, ValidationError),
    (429, RateLimitError),
    (500, ServerError),
    (503, ServerError),
])
def test_raise_for_status_maps_codes(code: int, exc_type: type) -> None:
    with pytest.raises(exc_type) as exc_info:
        raise_for_status(code, {"error": {"message": "test error"}})
    assert exc_info.value.status_code == code


def test_raise_for_status_generic_4xx():
    with pytest.raises(ThreeCXError) as exc_info:
        raise_for_status(409, "conflict")
    assert exc_info.value.status_code == 409


def test_raise_for_status_no_raise_on_2xx():
    raise_for_status(200, {})   # must not raise
    raise_for_status(204, None)  # must not raise


# --- Exception attributes -----------------------------------------------------

def test_exception_carries_status_code():
    err = AuthenticationError("bad token", status_code=401, detail={"x": 1})
    assert err.status_code == 401
    assert err.detail == {"x": 1}
    assert str(err) == "bad token"


def test_exception_repr():
    err = NotFoundError("missing", status_code=404)
    assert "404" in repr(err)
    assert "NotFoundError" in repr(err)


def test_is_subclass_of_base():
    assert issubclass(AuthenticationError, ThreeCXError)
    assert issubclass(NotFoundError, ThreeCXError)
    assert issubclass(ServerError, ThreeCXError)



"""Shared fixtures for the 3CX SDK test suite."""
from __future__ import annotations

import pytest
from pytest_httpx import HTTPXMock

from threecx import ThreeCXClient

BASE_URL = "https://pbx.test"
TOKEN_URL = f"{BASE_URL}/connect/token"
API_BASE = f"{BASE_URL}/xapi/v1"

TOKEN_RESPONSE = {"access_token": "test-token-abc", "expires_in": 3600, "token_type": "Bearer"}
TOKEN_JSON = {"access_token": "tok", "expires_in": 3600}


def api(path: str) -> str:
    """Build a full API URL from a relative path."""
    return API_BASE + path


@pytest.fixture
def mock_token(httpx_mock: HTTPXMock) -> None:
    """Pre-register a successful token response for every test that needs auth."""
    httpx_mock.add_response(
        url=TOKEN_URL,
        method="POST",
        json=TOKEN_RESPONSE,
    )


@pytest.fixture
def client(mock_token: None, httpx_mock: HTTPXMock) -> ThreeCXClient:  # noqa: ARG001
    """A ThreeCXClient wired to the mock PBX base URL."""
    return ThreeCXClient(
        base_url=BASE_URL,
        client_id="test-client-id",
        client_secret="test-client-secret",
    )



from __future__ import annotations
from typing import Any


class ThreeCXError(Exception):
    """Base exception for all 3CX SDK errors."""

    def __init__(self, message: str, status_code: int | None = None, detail: Any = None) -> None:
        super().__init__(message)
        self.status_code = status_code
        self.detail = detail

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(status_code={self.status_code}, message={str(self)!r})"


class AuthenticationError(ThreeCXError):
    """Raised when OAuth2 token acquisition or refresh fails."""


class NotFoundError(ThreeCXError):
    """Raised on HTTP 404."""


class ValidationError(ThreeCXError):
    """Raised on HTTP 400 / 422."""


class ServerError(ThreeCXError):
    """Raised on HTTP 5xx."""


class RateLimitError(ThreeCXError):
    """Raised on HTTP 429."""


def raise_for_status(status_code: int, body: Any) -> None:
    message = _extract_message(body)
    if status_code == 401:
        raise AuthenticationError(message, status_code, body)
    if status_code == 404:
        raise NotFoundError(message, status_code, body)
    if status_code in (400, 422):
        raise ValidationError(message, status_code, body)
    if status_code == 429:
        raise RateLimitError(message, status_code, body)
    if status_code >= 500:
        raise ServerError(message, status_code, body)
    if status_code >= 400:
        raise ThreeCXError(message, status_code, body)


def _extract_message(body: Any) -> str:
    if isinstance(body, dict):
        err = body.get("error", body)
        if isinstance(err, dict):
            return err.get("message", str(body))
        return str(err)
    return str(body)



from __future__ import annotations

from typing import Any, List, Optional

from pydantic import Field

from .base import _Base


class Trunk(_Base):
    """Represents a SIP trunk (Pbx.Trunk)."""

    id: Optional[int] = Field(None, alias="Id")
    name: Optional[str] = Field(None, alias="Name")
    type: Optional[str] = Field(None, alias="Type")
    host: Optional[str] = Field(None, alias="Host")
    port: Optional[int] = Field(None, alias="Port")
    enabled: Optional[bool] = Field(None, alias="Enabled")
    simultaneous_calls: Optional[int] = Field(None, alias="SimultaneousCalls")
    registered: Optional[bool] = Field(None, alias="Registered")
    number: Optional[str] = Field(None, alias="Number")
    inbound_rules: Optional[List[Any]] = Field(None, alias="InboundRules")
    outbound_rules: Optional[List[Any]] = Field(None, alias="OutboundRules")


class Peer(_Base):
    """Represents a SIP peer / VoIP gateway (Pbx.Peer)."""

    id: Optional[int] = Field(None, alias="Id")
    name: Optional[str] = Field(None, alias="Name")
    host: Optional[str] = Field(None, alias="Host")
    port: Optional[int] = Field(None, alias="Port")
    enabled: Optional[bool] = Field(None, alias="Enabled")


class Sbc(_Base):
    """Represents a Session Border Controller (Pbx.Sbc)."""

    id: Optional[int] = Field(None, alias="Id")
    name: Optional[str] = Field(None, alias="Name")
    host: Optional[str] = Field(None, alias="Host")
    port: Optional[int] = Field(None, alias="Port")
    enabled: Optional[bool] = Field(None, alias="Enabled")



from __future__ import annotations

from typing import Any, Dict, Iterator, List, Optional

from ..models.trunks import Peer, Sbc, Trunk
from ..odata import ODataQuery
from .base import BaseService


class TrunksService(BaseService):
    """CRUD operations for trunks, peers, and SBCs.

    Reference paths:
      GET    /Trunks
      POST   /Trunks
      GET    /Trunks({Id})
      PATCH  /Trunks({Id})
      DELETE /Trunks({Id})
      GET    /Peers
      POST   /Peers
      GET    /Peers({Id})
      PATCH  /Peers({Id})
      DELETE /Peers({Id})
      GET    /Sbcs
    """

    _TRUNKS = "/Trunks"
    _PEERS = "/Peers"
    _SBCS = "/Sbcs"

    # ------------------------------------------------------------------
    # Trunks
    # ------------------------------------------------------------------

    def list_trunks(self, query: Optional[ODataQuery] = None) -> List[Trunk]:
        data = self._list_raw(self._TRUNKS, query)
        return [Trunk.model_validate(item) for item in data.get("value", [])]

    def iterate_trunks(self, query: Optional[ODataQuery] = None) -> Iterator[Trunk]:
        yield from self._paginate(self._TRUNKS, Trunk, query)

    def create_trunk(self, trunk: Trunk | Dict[str, Any]) -> Trunk:
        payload = trunk.model_dump(by_alias=True, exclude_none=True) if isinstance(trunk, Trunk) else trunk
        data = self._post(self._TRUNKS, json=payload)
        return Trunk.model_validate(data)

    def get_trunk(self, trunk_id: int, query: Optional[ODataQuery] = None) -> Trunk:
        data = self._get(f"{self._TRUNKS}({trunk_id})", params=self._query_params(query))
        return Trunk.model_validate(data)

    def update_trunk(self, trunk_id: int, changes: Trunk | Dict[str, Any]) -> None:
        payload = changes.model_dump(by_alias=True, exclude_none=True) if isinstance(changes, Trunk) else changes
        self._patch(f"{self._TRUNKS}({trunk_id})", json=payload)

    def delete_trunk(self, trunk_id: int, etag: Optional[str] = None) -> None:
        self._delete(f"{self._TRUNKS}({trunk_id})", etag=etag)

    # ------------------------------------------------------------------
    # Peers
    # ------------------------------------------------------------------

    def list_peers(self, query: Optional[ODataQuery] = None) -> List[Peer]:
        data = self._list_raw(self._PEERS, query)
        return [Peer.model_validate(item) for item in data.get("value", [])]

    def get_peer(self, peer_id: int, query: Optional[ODataQuery] = None) -> Peer:
        data = self._get(f"{self._PEERS}({peer_id})", params=self._query_params(query))
        return Peer.model_validate(data)

    def create_peer(self, peer: Peer | Dict[str, Any]) -> Peer:
        payload = peer.model_dump(by_alias=True, exclude_none=True) if isinstance(peer, Peer) else peer
        data = self._post(self._PEERS, json=payload)
        return Peer.model_validate(data)

    def update_peer(self, peer_id: int, changes: Peer | Dict[str, Any]) -> None:
        payload = changes.model_dump(by_alias=True, exclude_none=True) if isinstance(changes, Peer) else changes
        self._patch(f"{self._PEERS}({peer_id})", json=payload)

    def delete_peer(self, peer_id: int, etag: Optional[str] = None) -> None:
        self._delete(f"{self._PEERS}({peer_id})", etag=etag)

    # ------------------------------------------------------------------
    # SBCs (read-only in most deployments)
    # ------------------------------------------------------------------

    def list_sbcs(self, query: Optional[ODataQuery] = None) -> List[Sbc]:
        data = self._list_raw(self._SBCS, query)
        return [Sbc.model_validate(item) for item in data.get("value", [])]

    def get_sbc(self, sbc_id: int, query: Optional[ODataQuery] = None) -> Sbc:
        data = self._get(f"{self._SBCS}({sbc_id})", params=self._query_params(query))
        return Sbc.model_validate(data)



"""Tests for ReportsService — URL construction and response parsing."""
from __future__ import annotations

import re
from datetime import datetime, timezone

import pytest
from pytest_httpx import HTTPXMock

from tests.conftest import TOKEN_URL, TOKEN_JSON, API_BASE
from threecx import ThreeCXClient
from threecx.services.reports import _fmt_dt


START = datetime(2026, 1, 1, 0, 0, 0, tzinfo=timezone.utc)
END = datetime(2026, 1, 7, 23, 59, 59, tzinfo=timezone.utc)

START_STR = "2026-01-01T00:00:00+00:00"
END_STR = "2026-01-07T23:59:59+00:00"


@pytest.fixture
def client(httpx_mock: HTTPXMock) -> ThreeCXClient:
    httpx_mock.add_response(url=TOKEN_URL, method="POST", json=TOKEN_JSON, is_reusable=True)
    return ThreeCXClient("https://pbx.test", "id", "secret")


# --- _fmt_dt helper -----------------------------------------------------------

def test_fmt_dt_formats_correctly() -> None:
    dt = datetime(2026, 3, 15, 9, 5, 30, tzinfo=timezone.utc)
    assert _fmt_dt(dt) == "2026-03-15T09:05:30+00:00"


# --- get_call_log -------------------------------------------------------------

def test_get_call_log(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    expected_path = (
        f"/ReportCallLogData/Pbx.GetCallLogData("
        f"periodFrom={START_STR},periodTo={END_STR})"
    )
    httpx_mock.add_response(
        url=API_BASE + expected_path,
        json={"value": [{"Id": "cdr-1", "Caller": "100", "TalkDuration": 60}]},
    )
    rows = client.reports.get_call_log(START, END)
    assert len(rows) == 1
    assert rows[0]["Caller"] == "100"


def test_get_call_log_empty(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    expected_path = (
        f"/ReportCallLogData/Pbx.GetCallLogData("
        f"periodFrom={START_STR},periodTo={END_STR})"
    )
    httpx_mock.add_response(url=API_BASE + expected_path, json={"value": []})
    rows = client.reports.get_call_log(START, END)
    assert rows == []


# --- get_extension_statistics -------------------------------------------------

def test_get_extension_statistics(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    path = (
        f"/ReportExtensionStatistics/Pbx.GetExtensionStatistics("
        f"periodFrom={START_STR},periodTo={END_STR})"
    )
    httpx_mock.add_response(
        url=API_BASE + path,
        json={"value": [{"Extension": "100", "TotalCalls": 15}]},
    )
    rows = client.reports.get_extension_statistics(START, END)
    assert rows[0]["TotalCalls"] == 15


# --- get_queue_performance_overview ------------------------------------------

def test_get_queue_performance_overview(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    path = (
        f"/ReportQueuePerformanceOverview/Pbx.GetQueuePerformanceOverview("
        f"periodFrom={START_STR},periodTo={END_STR})"
    )
    httpx_mock.add_response(
        url=API_BASE + path,
        json={"value": [{"QueueName": "Support", "AnsweredCalls": 50}]},
    )
    rows = client.reports.get_queue_performance_overview(START, END)
    assert rows[0]["QueueName"] == "Support"


# --- get_agent_login_history --------------------------------------------------

def test_get_agent_login_history(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    path = (
        f"/ReportAgentLoginHistory/Pbx.GetAgentLoginHistory("
        f"periodFrom={START_STR},periodTo={END_STR})"
    )
    httpx_mock.add_response(
        url=API_BASE + path,
        json={"value": [{"Agent": "100", "LoginTime": "2026-01-02T08:00:00Z"}]},
    )
    rows = client.reports.get_agent_login_history(START, END)
    assert rows[0]["Agent"] == "100"


# --- activity log -------------------------------------------------------------

def test_get_activity_log_filter(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=API_BASE + "/ActivityLog/Pbx.GetFilter()",
        json={"Extensions": ["100", "101"], "Severities": ["Info", "Warning"]},
    )
    result = client.reports.get_activity_log_filter()
    assert "Extensions" in result


def test_get_activity_logs_no_filters(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    path = (
        f"/ActivityLog/Pbx.GetLogs("
        f"startDate={START_STR},endDate={END_STR},"
        f"extension=null,call=null,severity=null)"
    )
    httpx_mock.add_response(
        url=API_BASE + path,
        json={"value": [{"Message": "System started", "Severity": "Info"}]},
    )
    rows = client.reports.get_activity_logs(START, END)
    assert rows[0]["Message"] == "System started"


def test_get_activity_logs_with_filters(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    path = (
        f"/ActivityLog/Pbx.GetLogs("
        f"startDate={START_STR},endDate={END_STR},"
        f"extension='100',call=null,severity='Warning')"
    )
    httpx_mock.add_response(
        url=API_BASE + path,
        json={"value": [{"Message": "Missed call", "Extension": "100"}]},
    )
    rows = client.reports.get_activity_logs(START, END, extension="100", severity="Warning")
    assert rows[0]["Extension"] == "100"


def test_purge_activity_logs(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=API_BASE + "/ActivityLog/Pbx.PurgeLogs",
        method="POST",
        status_code=204,
    )
    client.reports.purge_activity_logs()


# --- get_abandoned_queue_calls ------------------------------------------------

def test_get_abandoned_queue_calls(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    path = (
        f"/ReportAbandonedQueueCalls/Pbx.GetAbandonedQueueCalls("
        f"periodFrom={START_STR},periodTo={END_STR})"
    )
    httpx_mock.add_response(
        url=API_BASE + path,
        json={"value": [{"Caller": "0031612345678", "QueueName": "Support"}]},
    )
    rows = client.reports.get_abandoned_queue_calls(START, END)
    assert rows[0]["QueueName"] == "Support"



from __future__ import annotations

from typing import Any, List, Optional

from pydantic import Field

from .base import _Base


class Group(_Base):
    """A user group reference (Pbx.UserGroup)."""

    id: Optional[int] = Field(None, alias="Id")
    name: Optional[str] = Field(None, alias="Name")


class ForwardingProfile(_Base):
    """A forwarding profile entry (Pbx.ForwardingProfile)."""

    id: Optional[int] = Field(None, alias="Id")
    name: Optional[str] = Field(None, alias="Name")
    enabled: Optional[bool] = Field(None, alias="Enabled")


class Greeting(_Base):
    """A voicemail/IVR greeting file (Pbx.Greeting)."""

    id: Optional[int] = Field(None, alias="Id")
    filename: Optional[str] = Field(None, alias="GreetingFile")
    type: Optional[str] = Field(None, alias="GreetingType")
    name: Optional[str] = Field(None, alias="Name")


class User(_Base):
    """Represents a 3CX PBX user/extension (Pbx.User)."""

    id: Optional[int] = Field(None, alias="Id")
    number: Optional[str] = Field(None, alias="Number")
    first_name: Optional[str] = Field(None, alias="FirstName")
    last_name: Optional[str] = Field(None, alias="LastName")
    email: Optional[str] = Field(None, alias="Email")
    mobile: Optional[str] = Field(None, alias="Mobile")
    external_caller_id: Optional[str] = Field(None, alias="ExternalCallerID")
    internal_caller_id: Optional[str] = Field(None, alias="InternalCallerID")
    enabled: Optional[bool] = Field(None, alias="Enabled")
    is_registered: Optional[bool] = Field(None, alias="IsRegistered")
    status: Optional[str] = Field(None, alias="PresenceStatus")
    web_meeting_friendly_name: Optional[str] = Field(None, alias="WebMeetingFriendlyName")
    groups: Optional[List[Group]] = Field(None, alias="Groups")
    forwarding_profiles: Optional[List[ForwardingProfile]] = Field(None, alias="ForwardingProfiles")
    greetings: Optional[List[Greeting]] = Field(None, alias="Greetings")
    out_of_office_message: Optional[str] = Field(None, alias="OutOfOfficeMessage")
    language: Optional[str] = Field(None, alias="Language")
    primary_group_id: Optional[int] = Field(None, alias="PrimaryGroupId")
    transcription_mode: Optional[str] = Field(None, alias="TranscriptionMode")

    @property
    def full_name(self) -> str:
        parts = filter(None, [self.first_name, self.last_name])
        return " ".join(parts)



from __future__ import annotations

from typing import Optional

from pydantic import Field

from .base import _Base


class Phone(_Base):
    """Represents a provisioned IP phone (Pbx.Phone)."""

    id: Optional[int] = Field(None, alias="Id")
    mac: Optional[str] = Field(None, alias="MAC")
    model: Optional[str] = Field(None, alias="Model")
    ip_address: Optional[str] = Field(None, alias="IPAddress")
    firmware_version: Optional[str] = Field(None, alias="FirmwareVersion")
    assigned_extension: Optional[str] = Field(None, alias="AssignedExtension")
    last_registration: Optional[str] = Field(None, alias="LastRegistration")
    active: Optional[bool] = Field(None, alias="Active")


class PhoneTemplate(_Base):
    """Represents a phone provisioning template (Pbx.PhoneTemplate)."""

    id: Optional[int] = Field(None, alias="Id")
    name: Optional[str] = Field(None, alias="Name")
    model: Optional[str] = Field(None, alias="Model")
    manufacturer: Optional[str] = Field(None, alias="Manufacturer")



from __future__ import annotations

from typing import Any, List, Optional

from pydantic import Field

from .base import _Base


class QueueAgent(_Base):
    """An agent assigned to a queue (Pbx.QueueAgent)."""

    dn_number: Optional[str] = Field(None, alias="DnNumber")
    position: Optional[int] = Field(None, alias="Position")
    wrap_up_time: Optional[int] = Field(None, alias="WrapUpTime")


class QueueManager(_Base):
    """A manager assigned to a queue (Pbx.QueueManager)."""

    dn_number: Optional[str] = Field(None, alias="DnNumber")


class Queue(_Base):
    """Represents a 3CX call queue (Pbx.Queue)."""

    id: Optional[int] = Field(None, alias="Id")
    number: Optional[str] = Field(None, alias="Number")
    name: Optional[str] = Field(None, alias="Name")
    is_registered: Optional[bool] = Field(None, alias="IsRegistered")
    polling_strategy: Optional[str] = Field(None, alias="PollingStrategy")
    ring_timeout: Optional[int] = Field(None, alias="RingTimeout")
    master_timeout: Optional[int] = Field(None, alias="MasterTimeout")
    sla_time: Optional[int] = Field(None, alias="SLATime")
    announce_queue_position: Optional[bool] = Field(None, alias="AnnounceQueuePosition")
    announce_estimated_wait_time: Optional[bool] = Field(None, alias="AnnounceEstimatedWaitTime")
    agents: Optional[List[QueueAgent]] = Field(None, alias="Agents")
    managers: Optional[List[QueueManager]] = Field(None, alias="Managers")


class RingGroupMember(_Base):
    """A member of a ring group (Pbx.RingGroupMember)."""

    dn_number: Optional[str] = Field(None, alias="DnNumber")
    timeout: Optional[int] = Field(None, alias="Timeout")


class RingGroup(_Base):
    """Represents a 3CX ring group (Pbx.RingGroup)."""

    id: Optional[int] = Field(None, alias="Id")
    number: Optional[str] = Field(None, alias="Number")
    name: Optional[str] = Field(None, alias="Name")
    is_registered: Optional[bool] = Field(None, alias="IsRegistered")
    ring_strategy: Optional[str] = Field(None, alias="RingStrategy")
    ring_time: Optional[int] = Field(None, alias="RingTime")
    members: Optional[List[RingGroupMember]] = Field(None, alias="Members")
    greeting_file: Optional[str] = Field(None, alias="GreetingFile")



from __future__ import annotations

from typing import Optional

from pydantic import Field

from .base import _Base


class Contact(_Base):
    """Represents a PBX contact book entry (Pbx.Contact)."""

    id: Optional[int] = Field(None, alias="Id")
    first_name: Optional[str] = Field(None, alias="FirstName")
    last_name: Optional[str] = Field(None, alias="LastName")
    company_name: Optional[str] = Field(None, alias="CompanyName")
    email: Optional[str] = Field(None, alias="Email")
    phone_number: Optional[str] = Field(None, alias="PhoneNumber")
    mobile: Optional[str] = Field(None, alias="Mobile2")
    home: Optional[str] = Field(None, alias="Home")
    business: Optional[str] = Field(None, alias="Business")
    title: Optional[str] = Field(None, alias="Title")
    department: Optional[str] = Field(None, alias="Department")
    tag: Optional[str] = Field(None, alias="Tag")
    contact_type: Optional[str] = Field(None, alias="ContactType")

    @property
    def full_name(self) -> str:
        parts = filter(None, [self.first_name, self.last_name])
        return " ".join(parts)



from __future__ import annotations

from typing import Any, Dict, Iterator, List, Optional

from ..models.users import ForwardingProfile, Greeting, Group, User
from ..odata import ODataQuery
from .base import BaseService


class UsersService(BaseService):
    """CRUD operations for PBX users/extensions.

    Reference paths:
      GET    /Users
      POST   /Users
      GET    /Users({Id})
      PATCH  /Users({Id})
      DELETE /Users({Id})
      GET    /Users({Id})/Groups
      GET    /Users({Id})/ForwardingProfiles
      GET    /Users({Id})/Greetings
      POST   /Users({Id})/Pbx.SendWelcomeEmail
      POST   /Users/Pbx.BatchDelete
      POST   /Users/Pbx.MakeCall
    """

    _PATH = "/Users"

    # ------------------------------------------------------------------
    # Collection
    # ------------------------------------------------------------------

    def list(self, query: Optional[ODataQuery] = None) -> List[User]:
        """Return all users (single page)."""
        data = self._list_raw(self._PATH, query)
        return [User.model_validate(item) for item in data.get("value", [])]

    def iterate(self, query: Optional[ODataQuery] = None) -> Iterator[User]:
        """Yield users across all OData pages."""
        yield from self._paginate(self._PATH, User, query)

    def create(self, user: User | Dict[str, Any]) -> User:
        """Create a new user/extension."""
        payload = user.model_dump(by_alias=True, exclude_none=True) if isinstance(user, User) else user
        data = self._post(self._PATH, json=payload)
        return User.model_validate(data)

    # ------------------------------------------------------------------
    # Single entity
    # ------------------------------------------------------------------

    def get(
        self,
        user_id: int,
        query: Optional[ODataQuery] = None,
    ) -> User:
        """Retrieve a user by numeric ID."""
        data = self._get(f"{self._PATH}({user_id})", params=self._query_params(query))
        return User.model_validate(data)

    def update(self, user_id: int, changes: User | Dict[str, Any]) -> None:
        """Partially update a user (PATCH)."""
        payload = changes.model_dump(by_alias=True, exclude_none=True) if isinstance(changes, User) else changes
        self._patch(f"{self._PATH}({user_id})", json=payload)

    def delete(self, user_id: int, etag: Optional[str] = None) -> None:
        """Delete a user by ID. Optionally pass an ETag for optimistic concurrency."""
        self._delete(f"{self._PATH}({user_id})", etag=etag)

    # ------------------------------------------------------------------
    # Sub-resources
    # ------------------------------------------------------------------

    def get_groups(self, user_id: int) -> List[Group]:
        data = self._get(f"{self._PATH}({user_id})/Groups")
        return [Group.model_validate(item) for item in data.get("value", [])]

    def get_forwarding_profiles(self, user_id: int) -> List[ForwardingProfile]:
        data = self._get(f"{self._PATH}({user_id})/ForwardingProfiles")
        return [ForwardingProfile.model_validate(item) for item in data.get("value", [])]

    def get_greetings(self, user_id: int) -> List[Greeting]:
        data = self._get(f"{self._PATH}({user_id})/Greetings")
        return [Greeting.model_validate(item) for item in data.get("value", [])]

    # ------------------------------------------------------------------
    # Actions
    # ------------------------------------------------------------------

    def send_welcome_email(self, user_id: int) -> None:
        """Send the welcome e-mail to a user."""
        self._post(f"{self._PATH}({user_id})/Pbx.SendWelcomeEmail")

    def make_call(self, user_id: int, destination: str) -> None:
        """Initiate a click-to-call for the user towards *destination*."""
        self._post(f"{self._PATH}({user_id})/Pbx.MakeCall", json={"Destination": destination})

    def batch_delete(self, user_ids: List[int]) -> None:
        """Delete multiple users in one request."""
        self._post(f"{self._PATH}/Pbx.BatchDelete", json={"userIds": user_ids})

    def regenerate_passwords(self) -> None:
        """Trigger password regeneration for all users."""
        self._post(f"{self._PATH}/Pbx.RegeneratePasswords")

    def get_first_available_extension(self) -> str:
        """Return the next free extension number."""
        data = self._get(f"{self._PATH}/Pbx.GetFirstAvailableExtensionNumber()")
        return data.get("value", data)



from __future__ import annotations

from datetime import datetime
from typing import Optional

from pydantic import Field

from .base import _Base


class ActiveCall(_Base):
    """Represents a currently active call (Pbx.ActiveCall)."""

    id: Optional[int] = Field(None, alias="Id")
    status: Optional[str] = Field(None, alias="Status")
    established: Optional[bool] = Field(None, alias="Established")
    start_time: Optional[datetime] = Field(None, alias="StartTime")
    caller: Optional[str] = Field(None, alias="Caller")
    callee: Optional[str] = Field(None, alias="Callee")
    caller_display_name: Optional[str] = Field(None, alias="CallerDisplayName")
    callee_display_name: Optional[str] = Field(None, alias="CalleeDisplayName")
    duration: Optional[int] = Field(None, alias="Duration")
    on_hold: Optional[bool] = Field(None, alias="OnHold")
    recording: Optional[bool] = Field(None, alias="Recording")
    is_internal: Optional[bool] = Field(None, alias="IsInternal")


class CallHistoryEntry(_Base):
    """Represents an entry in the call history log (Pbx.CallLogData)."""

    id: Optional[str] = Field(None, alias="Id")
    caller: Optional[str] = Field(None, alias="Caller")
    callee: Optional[str] = Field(None, alias="Callee")
    caller_display_name: Optional[str] = Field(None, alias="CallerDisplayName")
    callee_display_name: Optional[str] = Field(None, alias="CalleeDisplayName")
    start_time: Optional[datetime] = Field(None, alias="StartTime")
    talk_duration: Optional[int] = Field(None, alias="TalkDuration")
    ring_duration: Optional[int] = Field(None, alias="RingDuration")
    status: Optional[str] = Field(None, alias="Status")
    direction: Optional[str] = Field(None, alias="CallDirection")
    recording_file: Optional[str] = Field(None, alias="RecordingFile")
    from_display_name: Optional[str] = Field(None, alias="From")
    to_display_name: Optional[str] = Field(None, alias="To")


class OutboundCall(_Base):
    """Body for initiating an outbound call (Pbx.OutboundCall)."""

    number: Optional[str] = Field(None, alias="Number")
    extension: Optional[str] = Field(None, alias="Extension")



"""Tests for OAuth2Auth token acquisition, caching, and invalidation."""
from __future__ import annotations

import time

import pytest
from pytest_httpx import HTTPXMock

from threecx.auth import OAuth2Auth, _Token
from threecx.exceptions import AuthenticationError

TOKEN_URL = "https://pbx.test/connect/token"


def _make_auth() -> OAuth2Auth:
    return OAuth2Auth(
        base_url="https://pbx.test",
        client_id="client-id",
        client_secret="client-secret",
    )


def test_fetches_token_on_first_use(httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=TOKEN_URL,
        method="POST",
        json={"access_token": "tok-abc", "expires_in": 3600},
    )
    auth = _make_auth()
    token = auth._get_token()
    assert token == "tok-abc"


def test_token_is_cached_after_first_fetch(httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=TOKEN_URL,
        method="POST",
        json={"access_token": "tok-xyz", "expires_in": 3600},
        is_reusable=False,
    )
    auth = _make_auth()
    auth._get_token()
    # Second call must NOT trigger another HTTP request
    token2 = auth._get_token()
    assert token2 == "tok-xyz"


def test_token_is_refreshed_when_expired(httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=TOKEN_URL,
        method="POST",
        json={"access_token": "tok-old", "expires_in": 3600},
    )
    auth = _make_auth()
    auth._get_token()

    # Force expiry
    auth._token = _Token(access_token="tok-old", expires_at=time.monotonic() - 1)

    httpx_mock.add_response(
        url=TOKEN_URL,
        method="POST",
        json={"access_token": "tok-new", "expires_in": 3600},
    )
    token = auth._get_token()
    assert token == "tok-new"


def test_invalidate_forces_refetch(httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=TOKEN_URL,
        method="POST",
        json={"access_token": "tok-1", "expires_in": 3600},
    )
    auth = _make_auth()
    auth._get_token()
    auth.invalidate()
    assert auth._token is None

    httpx_mock.add_response(
        url=TOKEN_URL,
        method="POST",
        json={"access_token": "tok-2", "expires_in": 3600},
    )
    token = auth._get_token()
    assert token == "tok-2"


def test_raises_authentication_error_on_non_200(httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=TOKEN_URL,
        method="POST",
        status_code=401,
        text="Unauthorized",
    )
    auth = _make_auth()
    with pytest.raises(AuthenticationError) as exc_info:
        auth._get_token()
    assert exc_info.value.status_code == 401


def test_raises_authentication_error_when_no_access_token(httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=TOKEN_URL,
        method="POST",
        json={"error": "invalid_client"},
    )
    auth = _make_auth()
    with pytest.raises(AuthenticationError, match="No access_token"):
        auth._get_token()


def test_auth_flow_sets_bearer_header(httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=TOKEN_URL,
        method="POST",
        json={"access_token": "bearer-tok", "expires_in": 3600},
    )
    import httpx as _httpx

    auth = _make_auth()
    request = _httpx.Request("GET", "https://pbx.test/xapi/v1/Users")
    # Consume the auth_flow generator
    gen = auth.auth_flow(request)
    next(gen)
    try:
        next(gen)
    except StopIteration:
        pass
    assert request.headers["Authorization"] == "Bearer bearer-tok"


def test_token_expiry_set_30s_before_actual_expiry(httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=TOKEN_URL,
        method="POST",
        json={"access_token": "tok", "expires_in": 3600},
    )
    auth = _make_auth()
    before = time.monotonic()
    auth._get_token()
    after = time.monotonic()
    # expires_at should be ~3570s from now (3600 - 30)
    assert before + 3570 - 1 <= auth._token.expires_at <= after + 3570 + 1



"""Tests for QueuesService."""
from __future__ import annotations

import pytest
from pytest_httpx import HTTPXMock

from tests.conftest import TOKEN_URL, TOKEN_JSON, api
from threecx import ThreeCXClient
from threecx.models.queues import Queue


QUEUE_1 = {"Id": 10, "Number": "800", "Name": "Support", "RingTimeout": 30, "SLATime": 20}
QUEUE_2 = {"Id": 11, "Number": "801", "Name": "Sales", "RingTimeout": 20}


@pytest.fixture
def client(httpx_mock: HTTPXMock) -> ThreeCXClient:
    httpx_mock.add_response(url=TOKEN_URL, method="POST", json=TOKEN_JSON, is_reusable=True)
    return ThreeCXClient("https://pbx.test", "id", "secret")


def test_list_queues(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(url=api("/Queues"), json={"value": [QUEUE_1, QUEUE_2]})
    queues = client.queues.list()
    assert len(queues) == 2
    assert queues[0].name == "Support"


def test_get_queue_by_id(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(url=api("/Queues(10)"), json=QUEUE_1)
    q = client.queues.get(10)
    assert q.ring_timeout == 30


def test_get_queue_by_number(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(url=api("/Queues(Number='800')"), json=QUEUE_1)
    q = client.queues.get_by_number("800")
    assert q.id == 10


def test_create_queue(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=api("/Queues"),
        method="POST",
        status_code=201,
        json={**QUEUE_1, "Id": 12},
    )
    created = client.queues.create(Queue(Number="802", Name="Billing", RingTimeout=25))
    assert created.id == 12


def test_update_queue(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(url=api("/Queues(10)"), method="PATCH", status_code=204)
    client.queues.update(10, {"SLATime": 30})


def test_delete_queue(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(url=api("/Queues(10)"), method="DELETE", status_code=204)
    client.queues.delete(10)


def test_get_agents(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=api("/Queues(10)/Agents"),
        json={"value": [{"DnNumber": "100", "Position": 1}, {"DnNumber": "101", "Position": 2}]},
    )
    agents = client.queues.get_agents(10)
    assert len(agents) == 2
    assert agents[0].dn_number == "100"


def test_get_managers(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=api("/Queues(10)/Managers"),
        json={"value": [{"DnNumber": "102"}]},
    )
    managers = client.queues.get_managers(10)
    assert managers[0].dn_number == "102"


def test_reset_statistics(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=api("/Queues(10)/Pbx.ResetQueueStatistics"),
        method="POST",
        status_code=204,
    )
    client.queues.reset_statistics(10)



from __future__ import annotations

from typing import Any, Dict, Optional
from urllib.parse import urljoin

import httpx

from .auth import OAuth2Auth
from .services.active_calls import ActiveCallsService
from .services.call_history import CallHistoryService
from .services.contacts import ContactsService
from .services.phones import PhonesService
from .services.queues import QueuesService
from .services.reports import ReportsService
from .services.ring_groups import RingGroupsService
from .services.system import SystemService
from .services.trunks import TrunksService
from .services.users import UsersService

_DEFAULT_API_PATH = "/xapi/v1"


class ThreeCXClient:
    """Entry point for the 3CX XAPI Python SDK.

    Usage::

        client = ThreeCXClient(
            base_url="https://pbx.example.com",
            client_id="my-service-principal-id",
            client_secret="my-secret",
        )

        # List active calls
        for call in client.active_calls.list():
            print(call.caller, "→", call.callee)

        # Find a user by extension number
        users = client.users.list(
            ODataQuery().filter("Number eq '100'")
        )

        # Pull 7 days of call history
        from datetime import datetime, timedelta
        end = datetime.utcnow()
        start = end - timedelta(days=7)
        entries = client.reports.get_call_log(start, end)

    Args:
        base_url: Root URL of the 3CX instance, e.g. ``https://pbx.example.com``.
        client_id: Service-principal client ID (Admin → Integrations → API).
        client_secret: Service-principal client secret.
        api_path: API prefix; defaults to ``/xapi/v1``.
        timeout: HTTP request timeout in seconds (default 30).
        verify_ssl: Set to ``False`` to skip TLS verification (not recommended in production).
        httpx_kwargs: Additional keyword arguments forwarded to :class:`httpx.Client`.
    """

    def __init__(
        self,
        base_url: str,
        client_id: str,
        client_secret: str,
        *,
        api_path: str = _DEFAULT_API_PATH,
        timeout: float = 30.0,
        verify_ssl: bool = True,
        **httpx_kwargs: Any,
    ) -> None:
        self._base_url = base_url.rstrip("/")
        self._auth = OAuth2Auth(
            base_url=self._base_url,
            client_id=client_id,
            client_secret=client_secret,
        )
        self._http = httpx.Client(
            base_url=self._base_url + api_path,
            auth=self._auth,
            timeout=timeout,
            verify=verify_ssl,
            headers={"Accept": "application/json", "Content-Type": "application/json"},
            **httpx_kwargs,
        )

        # --- service registry ---
        self.active_calls = ActiveCallsService(self._http)
        self.users = UsersService(self._http)
        self.queues = QueuesService(self._http)
        self.ring_groups = RingGroupsService(self._http)
        self.call_history = CallHistoryService(self._http)
        self.trunks = TrunksService(self._http)
        self.contacts = ContactsService(self._http)
        self.phones = PhonesService(self._http)
        self.system = SystemService(self._http)
        self.reports = ReportsService(self._http)

    # ------------------------------------------------------------------
    # Context-manager support
    # ------------------------------------------------------------------

    def __enter__(self) -> ThreeCXClient:
        return self

    def __exit__(self, *_: Any) -> None:
        self.close()

    def close(self) -> None:
        """Close the underlying HTTP session."""
        self._http.close()

    # ------------------------------------------------------------------
    # Convenience / escape hatches
    # ------------------------------------------------------------------

    def get(self, path: str, params: Optional[Dict[str, Any]] = None) -> Any:
        """Perform a raw GET against any XAPI path."""
        return self.system.get_raw(path, params=params)

    def post(self, path: str, json: Any = None) -> Any:
        """Perform a raw POST against any XAPI path."""
        return self.system.post_raw(path, json=json)

    def invalidate_token(self) -> None:
        """Force re-authentication on the next request."""
        self._auth.invalidate()

    def __repr__(self) -> str:
        return f"ThreeCXClient(base_url={self._base_url!r})"



from __future__ import annotations

from typing import Any, List, Optional

from pydantic import Field

from .base import _Base


class SystemStatus(_Base):
    """Represents the system status (Pbx.SystemStatus)."""

    has_active_calls: Optional[bool] = Field(None, alias="HasActiveCalls")
    calls_active: Optional[int] = Field(None, alias="CallsActive")
    extensions_registered: Optional[int] = Field(None, alias="ExtensionsRegistered")
    extensions_total: Optional[int] = Field(None, alias="ExtensionsTotal")
    trunks_registered: Optional[int] = Field(None, alias="TrunksRegistered")
    trunks_total: Optional[int] = Field(None, alias="TrunksTotal")
    disk_usage: Optional[float] = Field(None, alias="DiskUsage")
    memory_usage: Optional[float] = Field(None, alias="MemUsage")
    cpu_usage: Optional[float] = Field(None, alias="CpuUsage")
    free_space: Optional[int] = Field(None, alias="FreeSpace")
    ip: Optional[str] = Field(None, alias="Ip")
    fqdn: Optional[str] = Field(None, alias="FQDN")
    pbx_version: Optional[str] = Field(None, alias="Version")
    up_since: Optional[str] = Field(None, alias="UpSince")


class LicenseStatus(_Base):
    """Represents the license / subscription status (Pbx.LicenseStatus)."""

    product_code: Optional[str] = Field(None, alias="ProductCode")
    max_simultaneous_calls: Optional[int] = Field(None, alias="MaxSimCalls")
    max_extensions: Optional[int] = Field(None, alias="MaxExtensions")
    registered_extensions: Optional[int] = Field(None, alias="RegisteredExtensions")
    subscription_valid_until: Optional[str] = Field(None, alias="SubscriptionValidUntil")
    maintenance_expires: Optional[str] = Field(None, alias="MaintenanceExpires")


class SystemParameters(_Base):
    """Represents global PBX parameters (Pbx.Parameters)."""

    fqdn: Optional[str] = Field(None, alias="FQDN")
    name: Optional[str] = Field(None, alias="Name")
    ip: Optional[str] = Field(None, alias="Ip")
    http_port: Optional[int] = Field(None, alias="HttpPort")
    https_port: Optional[int] = Field(None, alias="HttpsPort")
    sip_port: Optional[int] = Field(None, alias="SIPPort")
    tunnel_port: Optional[int] = Field(None, alias="TunnelPort")
    local_ip_address: Optional[str] = Field(None, alias="LocalIpAddress")
    supported_timezones: Optional[List[Any]] = Field(None, alias="SupportedTimezones")



from __future__ import annotations

from typing import Any, Dict, Iterator, List, Optional

from ..models.contacts import Contact
from ..odata import ODataQuery
from .base import BaseService


class ContactsService(BaseService):
    """CRUD operations for the PBX contact book.

    Reference paths:
      GET    /Contacts
      POST   /Contacts
      GET    /Contacts({Id})
      PATCH  /Contacts({Id})
      DELETE /Contacts({Id})
    """

    _PATH = "/Contacts"

    def list(self, query: Optional[ODataQuery] = None) -> List[Contact]:
        data = self._list_raw(self._PATH, query)
        return [Contact.model_validate(item) for item in data.get("value", [])]

    def iterate(self, query: Optional[ODataQuery] = None) -> Iterator[Contact]:
        yield from self._paginate(self._PATH, Contact, query)

    def create(self, contact: Contact | Dict[str, Any]) -> Contact:
        payload = (
            contact.model_dump(by_alias=True, exclude_none=True)
            if isinstance(contact, Contact)
            else contact
        )
        data = self._post(self._PATH, json=payload)
        return Contact.model_validate(data)

    def get(self, contact_id: int, query: Optional[ODataQuery] = None) -> Contact:
        data = self._get(f"{self._PATH}({contact_id})", params=self._query_params(query))
        return Contact.model_validate(data)

    def update(self, contact_id: int, changes: Contact | Dict[str, Any]) -> None:
        payload = (
            changes.model_dump(by_alias=True, exclude_none=True)
            if isinstance(changes, Contact)
            else changes
        )
        self._patch(f"{self._PATH}({contact_id})", json=payload)

    def delete(self, contact_id: int, etag: Optional[str] = None) -> None:
        self._delete(f"{self._PATH}({contact_id})", etag=etag)



from .active_calls import ActiveCallsService
from .users import UsersService
from .queues import QueuesService
from .ring_groups import RingGroupsService
from .call_history import CallHistoryService
from .trunks import TrunksService
from .contacts import ContactsService
from .phones import PhonesService
from .system import SystemService
from .reports import ReportsService

__all__ = [
    "ActiveCallsService",
    "UsersService",
    "QueuesService",
    "RingGroupsService",
    "CallHistoryService",
    "TrunksService",
    "ContactsService",
    "PhonesService",
    "SystemService",
    "ReportsService",
]



from .client import ThreeCXClient
from .auth import OAuth2Auth
from .exceptions import (
    ThreeCXError,
    AuthenticationError,
    NotFoundError,
    ValidationError,
    ServerError,
)
from .odata import ODataQuery

__all__ = [
    "ThreeCXClient",
    "OAuth2Auth",
    "ThreeCXError",
    "AuthenticationError",
    "NotFoundError",
    "ValidationError",
    "ServerError",
    "ODataQuery",
]



from __future__ import annotations

from typing import Iterator, List, Optional

from ..models.calls import CallHistoryEntry
from ..odata import ODataQuery
from .base import BaseService


class CallHistoryService(BaseService):
    """Query the call history log.

    Reference paths:
      GET /CallHistoryView
    """

    _PATH = "/CallHistoryView"

    def list(self, query: Optional[ODataQuery] = None) -> List[CallHistoryEntry]:
        """Return a single page of call history entries."""
        data = self._list_raw(self._PATH, query)
        return [CallHistoryEntry.model_validate(item) for item in data.get("value", [])]

    def iterate(self, query: Optional[ODataQuery] = None) -> Iterator[CallHistoryEntry]:
        """Yield call history entries across all OData pages."""
        yield from self._paginate(self._PATH, CallHistoryEntry, query)

    def count(self, query: Optional[ODataQuery] = None) -> int:
        """Return the total number of call history records matching *query*."""
        q = (query or ODataQuery()).count().top(0)
        data = self._list_raw(self._PATH, q)
        return data.get("@odata.count", 0)



from __future__ import annotations

from typing import Any, Dict, Iterator, List, Optional, Type, TypeVar

import httpx

from ..exceptions import raise_for_status
from ..odata import ODataQuery

T = TypeVar("T")

_SENTINEL = object()


class BaseService:
    """Shared HTTP helpers used by every resource service."""

    def __init__(self, http: httpx.Client) -> None:
        self._http = http

    # ------------------------------------------------------------------
    # Low-level request helpers
    # ------------------------------------------------------------------

    def _get(self, path: str, params: Optional[Dict[str, Any]] = None) -> Any:
        response = self._http.get(path, params=params)
        self._raise(response)
        if response.status_code == 204:
            return None
        return response.json()

    def _post(self, path: str, json: Any = None, params: Optional[Dict[str, Any]] = None) -> Any:
        response = self._http.post(path, json=json, params=params)
        self._raise(response)
        if response.status_code == 204:
            return None
        return response.json()

    def _patch(self, path: str, json: Any) -> None:
        response = self._http.patch(path, json=json)
        self._raise(response)

    def _delete(self, path: str, etag: Optional[str] = None) -> None:
        headers = {"If-Match": etag} if etag else {}
        response = self._http.delete(path, headers=headers)
        self._raise(response)

    @staticmethod
    def _raise(response: httpx.Response) -> None:
        if response.is_error:
            try:
                body = response.json()
            except Exception:
                body = response.text
            raise_for_status(response.status_code, body)

    # ------------------------------------------------------------------
    # OData helpers
    # ------------------------------------------------------------------

    @staticmethod
    def _query_params(query: Optional[ODataQuery]) -> Dict[str, Any]:
        return query.to_params() if query else {}

    def _list_raw(self, path: str, query: Optional[ODataQuery] = None) -> Dict[str, Any]:
        return self._get(path, params=self._query_params(query))

    def _paginate(
        self,
        path: str,
        model: Type[T],
        query: Optional[ODataQuery] = None,
    ) -> Iterator[T]:
        """Yield all items across OData pages."""
        params: Optional[Dict[str, Any]] = self._query_params(query) or None
        while path:
            data = self._get(path, params=params)
            for item in data.get("value", []):
                yield model.model_validate(item)
            path = data.get("@odata.nextLink")
            # nextLink is a full URL with query string already embedded;
            # passing params={} to httpx strips the query string, so use None.
            params = None



"""Tests for TrunksService (trunks, peers, SBCs)."""
from __future__ import annotations

import pytest
from pytest_httpx import HTTPXMock

from tests.conftest import TOKEN_URL, TOKEN_JSON, api
from threecx import ThreeCXClient
from threecx.models.trunks import Trunk


TRUNK_1 = {"Id": 1, "Name": "SIP Carrier", "Host": "sip.carrier.com", "Port": 5060, "Enabled": True}
PEER_1 = {"Id": 1, "Name": "PSTN GW", "Host": "192.168.1.100", "Port": 5060, "Enabled": True}
SBC_1 = {"Id": 1, "Name": "SBC Edge", "Host": "10.0.0.1", "Port": 5060, "Enabled": True}


@pytest.fixture
def client(httpx_mock: HTTPXMock) -> ThreeCXClient:
    httpx_mock.add_response(url=TOKEN_URL, method="POST", json=TOKEN_JSON, is_reusable=True)
    return ThreeCXClient("https://pbx.test", "id", "secret")


# --- Trunks -------------------------------------------------------------------

def test_list_trunks(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(url=api("/Trunks"), json={"value": [TRUNK_1]})
    trunks = client.trunks.list_trunks()
    assert trunks[0].host == "sip.carrier.com"


def test_get_trunk(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(url=api("/Trunks(1)"), json=TRUNK_1)
    t = client.trunks.get_trunk(1)
    assert t.name == "SIP Carrier"


def test_create_trunk(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=api("/Trunks"),
        method="POST",
        status_code=201,
        json={**TRUNK_1, "Id": 2},
    )
    created = client.trunks.create_trunk(Trunk(Name="New Trunk", Host="sip2.carrier.com"))
    assert created.id == 2


def test_update_trunk(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(url=api("/Trunks(1)"), method="PATCH", status_code=204)
    client.trunks.update_trunk(1, {"Enabled": False})


def test_delete_trunk(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(url=api("/Trunks(1)"), method="DELETE", status_code=204)
    client.trunks.delete_trunk(1)


# --- Peers --------------------------------------------------------------------

def test_list_peers(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(url=api("/Peers"), json={"value": [PEER_1]})
    peers = client.trunks.list_peers()
    assert peers[0].host == "192.168.1.100"


def test_get_peer(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(url=api("/Peers(1)"), json=PEER_1)
    p = client.trunks.get_peer(1)
    assert p.name == "PSTN GW"


# --- SBCs ---------------------------------------------------------------------

def test_list_sbcs(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(url=api("/Sbcs"), json={"value": [SBC_1]})
    sbcs = client.trunks.list_sbcs()
    assert sbcs[0].name == "SBC Edge"


def test_get_sbc(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(url=api("/Sbcs(1)"), json=SBC_1)
    sbc = client.trunks.get_sbc(1)
    assert sbc.host == "10.0.0.1"



from __future__ import annotations

from typing import Iterator, List, Optional

from ..models.calls import ActiveCall
from ..odata import ODataQuery
from .base import BaseService


class ActiveCallsService(BaseService):
    """Manage and inspect currently active calls.

    Reference paths:
      GET  /ActiveCalls
      POST /ActiveCalls({Id})/Pbx.DropCall
    """

    _PATH = "/ActiveCalls"

    def list(self, query: Optional[ODataQuery] = None) -> List[ActiveCall]:
        """Return all active calls (single page)."""
        data = self._list_raw(self._PATH, query)
        return [ActiveCall.model_validate(item) for item in data.get("value", [])]

    def iterate(self, query: Optional[ODataQuery] = None) -> Iterator[ActiveCall]:
        """Yield active calls across all OData pages."""
        yield from self._paginate(self._PATH, ActiveCall, query)

    def drop(self, call_id: int) -> None:
        """Forcefully drop an active call by its ID."""
        self._post(f"{self._PATH}({call_id})/Pbx.DropCall")



"""Tests for ThreeCXClient construction, context manager, and escape hatches."""
from __future__ import annotations

import pytest
from pytest_httpx import HTTPXMock

from threecx import ThreeCXClient
from threecx.services.active_calls import ActiveCallsService
from threecx.services.call_history import CallHistoryService
from threecx.services.contacts import ContactsService
from threecx.services.phones import PhonesService
from threecx.services.queues import QueuesService
from threecx.services.reports import ReportsService
from threecx.services.ring_groups import RingGroupsService
from threecx.services.system import SystemService
from threecx.services.trunks import TrunksService
from threecx.services.users import UsersService

TOKEN_URL = "https://pbx.test/connect/token"
TOKEN_JSON = {"access_token": "tok", "expires_in": 3600}


@pytest.fixture
def c(httpx_mock: HTTPXMock) -> ThreeCXClient:
    # is_optional=True because some tests don't trigger an HTTP request at all
    # (e.g. repr, attribute checks) so the token endpoint may not be called.
    httpx_mock.add_response(url=TOKEN_URL, method="POST", json=TOKEN_JSON, is_reusable=True, is_optional=True)
    return ThreeCXClient("https://pbx.test", "id", "secret")


def test_all_services_registered(c: ThreeCXClient) -> None:
    assert isinstance(c.active_calls, ActiveCallsService)
    assert isinstance(c.users, UsersService)
    assert isinstance(c.queues, QueuesService)
    assert isinstance(c.ring_groups, RingGroupsService)
    assert isinstance(c.call_history, CallHistoryService)
    assert isinstance(c.trunks, TrunksService)
    assert isinstance(c.contacts, ContactsService)
    assert isinstance(c.phones, PhonesService)
    assert isinstance(c.system, SystemService)
    assert isinstance(c.reports, ReportsService)


def test_repr(c: ThreeCXClient) -> None:
    assert "https://pbx.test" in repr(c)


def test_context_manager(httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(url=TOKEN_URL, method="POST", json=TOKEN_JSON, is_optional=True)
    with ThreeCXClient("https://pbx.test", "id", "secret") as c:
        assert isinstance(c, ThreeCXClient)
    # __exit__ calls close(); if the client is already closed a second close
    # would raise — we just check it didn't raise here.


def test_escape_hatch_get(c: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url="https://pbx.test/xapi/v1/InboundRules",
        json={"value": [{"Id": 1}]},
    )
    data = c.get("/InboundRules")
    assert data["value"][0]["Id"] == 1


def test_escape_hatch_post(c: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url="https://pbx.test/xapi/v1/Backups/Pbx.CreateBackup",
        method="POST",
        status_code=204,
    )
    result = c.post("/Backups/Pbx.CreateBackup")
    assert result is None


def test_invalidate_token(c: ThreeCXClient) -> None:
    c.invalidate_token()
    assert c._auth._token is None



from __future__ import annotations

import time
from dataclasses import dataclass, field

import httpx

from .exceptions import AuthenticationError


@dataclass
class _Token:
    access_token: str
    expires_at: float


@dataclass
class OAuth2Auth(httpx.Auth):
    """OAuth2 client-credentials authenticator for the 3CX XAPI.

    Tokens are cached and refreshed automatically when they expire.

    Args:
        base_url: Base URL of the 3CX instance, e.g. ``https://pbx.example.com``.
        client_id: Service-principal client ID.
        client_secret: Service-principal client secret.
        token_path: Path to the token endpoint (default: ``/connect/token``).
    """

    base_url: str
    client_id: str
    client_secret: str
    token_path: str = "/connect/token"
    _token: _Token | None = field(default=None, init=False, repr=False)

    # --- httpx.Auth protocol --------------------------------------------------

    def auth_flow(self, request: httpx.Request):
        token = self._get_token()
        request.headers["Authorization"] = f"Bearer {token}"
        yield request

    # --- internal helpers -----------------------------------------------------

    def _get_token(self) -> str:
        if self._token is None or time.monotonic() >= self._token.expires_at:
            self._token = self._fetch_token()
        return self._token.access_token

    def _fetch_token(self) -> _Token:
        url = self.base_url.rstrip("/") + self.token_path
        try:
            response = httpx.post(
                url,
                data={
                    "grant_type": "client_credentials",
                    "client_id": self.client_id,
                    "client_secret": self.client_secret,
                },
                headers={"Content-Type": "application/x-www-form-urlencoded"},
            )
        except httpx.HTTPError as exc:
            raise AuthenticationError(f"Token request failed: {exc}") from exc

        if response.status_code != 200:
            raise AuthenticationError(
                f"Token endpoint returned {response.status_code}: {response.text}",
                response.status_code,
            )

        data = response.json()
        if "access_token" not in data:
            raise AuthenticationError(f"No access_token in response: {data}")

        expires_in = int(data.get("expires_in", 3600))
        # Subtract 30 s to refresh before actual expiry
        expires_at = time.monotonic() + expires_in - 30
        return _Token(access_token=data["access_token"], expires_at=expires_at)

    def invalidate(self) -> None:
        """Force the next request to re-acquire a token."""
        self._token = None



from .base import ODataCollection, ODataError
from .calls import ActiveCall, CallHistoryEntry, OutboundCall
from .users import User, Group, ForwardingProfile, Greeting
from .queues import Queue, QueueAgent, QueueManager, RingGroup, RingGroupMember
from .trunks import Trunk, Peer, Sbc
from .system import SystemStatus, LicenseStatus, SystemParameters
from .contacts import Contact
from .phones import Phone, PhoneTemplate

__all__ = [
    "ODataCollection",
    "ODataError",
    "ActiveCall",
    "CallHistoryEntry",
    "OutboundCall",
    "User",
    "Group",
    "ForwardingProfile",
    "Greeting",
    "Queue",
    "QueueAgent",
    "QueueManager",
    "RingGroup",
    "RingGroupMember",
    "Trunk",
    "Peer",
    "Sbc",
    "SystemStatus",
    "LicenseStatus",
    "SystemParameters",
    "Contact",
    "Phone",
    "PhoneTemplate",
]



"""Tests for UsersService — CRUD, sub-resources, actions, and OData queries."""
from __future__ import annotations

import pytest
from pytest_httpx import HTTPXMock

from tests.conftest import TOKEN_URL, TOKEN_JSON, api
from threecx import ThreeCXClient
from threecx.models.users import User
from threecx.odata import ODataQuery

USER_1 = {"Id": 101, "Number": "100", "FirstName": "Alice", "LastName": "Smith", "Email": "alice@x.com", "Enabled": True}
USER_2 = {"Id": 102, "Number": "101", "FirstName": "Bob", "LastName": "Jones", "Enabled": True}



@pytest.fixture
def client(httpx_mock: HTTPXMock) -> ThreeCXClient:
    httpx_mock.add_response(url=TOKEN_URL, method="POST", json=TOKEN_JSON, is_reusable=True)
    return ThreeCXClient("https://pbx.test", "id", "secret")


# --- list ---------------------------------------------------------------------

def test_list_returns_users(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=api("/Users"),
        json={"value": [USER_1, USER_2]},
    )
    users = client.users.list()
    assert len(users) == 2
    assert users[0].first_name == "Alice"
    assert users[1].number == "101"


def test_list_empty_collection(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(url=api("/Users"), json={"value": []})
    assert client.users.list() == []


def test_list_with_odata_query(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=api("/Users"),
        match_params={"$filter": "Enabled eq true", "$top": "5"},
        json={"value": [USER_1]},
    )
    q = ODataQuery().filter("Enabled eq true").top(5)
    users = client.users.list(q)
    assert len(users) == 1


# --- iterate (pagination) -----------------------------------------------------

def test_iterate_follows_next_link(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=api("/Users"),
        json={"value": [USER_1], "@odata.nextLink": api("/Users?$skip=1")},
    )
    httpx_mock.add_response(
        url=api("/Users?$skip=1"),
        json={"value": [USER_2]},
    )
    users = list(client.users.iterate())
    assert len(users) == 2
    assert users[0].id == 101
    assert users[1].id == 102


# --- get ----------------------------------------------------------------------

def test_get_user_by_id(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(url=api("/Users(101)"), json=USER_1)
    user = client.users.get(101)
    assert user.id == 101
    assert user.email == "alice@x.com"


def test_get_user_with_expand(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=api("/Users(101)"),
        match_params={"$expand": "Groups"},
        json={**USER_1, "Groups": [{"Id": 5, "Name": "Sales"}]},
    )
    user = client.users.get(101, query=ODataQuery().expand("Groups"))
    assert user.groups[0].name == "Sales"


# --- create -------------------------------------------------------------------

def test_create_user(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    new_payload = {"Number": "200", "FirstName": "Carol", "Email": "carol@x.com"}
    created_payload = {**new_payload, "Id": 103}
    httpx_mock.add_response(
        url=api("/Users"),
        method="POST",
        status_code=201,
        json=created_payload,
    )
    created = client.users.create(User(Number="200", FirstName="Carol", Email="carol@x.com"))
    assert created.id == 103
    assert created.number == "200"


def test_create_user_from_dict(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=api("/Users"),
        method="POST",
        status_code=201,
        json={"Id": 104, "Number": "201"},
    )
    created = client.users.create({"Number": "201", "FirstName": "Dave"})
    assert created.id == 104


# --- update -------------------------------------------------------------------

def test_update_user(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=api("/Users(101)"),
        method="PATCH",
        status_code=204,
    )
    client.users.update(101, {"Email": "newalice@x.com"})  # must not raise


# --- delete -------------------------------------------------------------------

def test_delete_user(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=api("/Users(101)"),
        method="DELETE",
        status_code=204,
    )
    client.users.delete(101)


def test_delete_user_with_etag(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=api("/Users(101)"),
        method="DELETE",
        status_code=204,
        match_headers={"If-Match": '"abc123"'},
    )
    client.users.delete(101, etag='"abc123"')


# --- sub-resources ------------------------------------------------------------

def test_get_groups(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=api("/Users(101)/Groups"),
        json={"value": [{"Id": 5, "Name": "Sales"}, {"Id": 6, "Name": "Support"}]},
    )
    groups = client.users.get_groups(101)
    assert len(groups) == 2
    assert groups[0].name == "Sales"


def test_get_forwarding_profiles(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=api("/Users(101)/ForwardingProfiles"),
        json={"value": [{"Id": 1, "Name": "Away", "Enabled": True}]},
    )
    profiles = client.users.get_forwarding_profiles(101)
    assert profiles[0].name == "Away"


def test_get_greetings(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=api("/Users(101)/Greetings"),
        json={"value": [{"Id": 1, "GreetingFile": "voicemail.wav", "GreetingType": "Voicemail"}]},
    )
    greetings = client.users.get_greetings(101)
    assert greetings[0].filename == "voicemail.wav"


# --- actions ------------------------------------------------------------------

def test_send_welcome_email(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=api("/Users(101)/Pbx.SendWelcomeEmail"),
        method="POST",
        status_code=204,
    )
    client.users.send_welcome_email(101)


def test_make_call(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=api("/Users(101)/Pbx.MakeCall"),
        method="POST",
        match_json={"Destination": "0031612345678"},
        status_code=204,
    )
    client.users.make_call(101, "0031612345678")


def test_batch_delete(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=api("/Users/Pbx.BatchDelete"),
        method="POST",
        match_json={"userIds": [101, 102]},
        status_code=204,
    )
    client.users.batch_delete([101, 102])


def test_get_first_available_extension(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=api("/Users/Pbx.GetFirstAvailableExtensionNumber()"),
        json={"value": "205"},
    )
    ext = client.users.get_first_available_extension()
    assert ext == "205"



"""Unit tests for ODataQuery — no HTTP required."""
from __future__ import annotations

import pytest

from threecx.odata import ODataQuery


def test_empty_query_produces_no_params():
    assert ODataQuery().to_params() == {}


def test_top():
    assert ODataQuery().top(25).to_params() == {"$top": 25}


def test_skip():
    assert ODataQuery().skip(50).to_params() == {"$skip": 50}


def test_count_default_true():
    assert ODataQuery().count().to_params() == {"$count": "true"}


def test_count_false():
    assert ODataQuery().count(False).to_params() == {"$count": "false"}


def test_select_single():
    assert ODataQuery().select("Id").to_params() == {"$select": "Id"}


def test_select_multiple():
    params = ODataQuery().select("Id", "FirstName", "LastName").to_params()
    assert params["$select"] == "Id,FirstName,LastName"


def test_expand():
    assert ODataQuery().expand("Groups", "Greetings").to_params() == {
        "$expand": "Groups,Greetings"
    }


def test_filter():
    params = ODataQuery().filter("Enabled eq true").to_params()
    assert params["$filter"] == "Enabled eq true"


def test_search():
    assert ODataQuery().search("alice").to_params() == {"$search": "alice"}


def test_order_by_single():
    assert ODataQuery().order_by("Number").to_params() == {"$orderby": "Number"}


def test_order_by_multiple():
    params = ODataQuery().order_by("LastName", "FirstName").to_params()
    assert params["$orderby"] == "LastName,FirstName"


def test_chaining_all_params():
    q = (
        ODataQuery()
        .filter("Enabled eq true")
        .select("Id", "Number")
        .expand("Groups")
        .order_by("Number")
        .top(10)
        .skip(20)
        .count()
    )
    params = q.to_params()
    assert params["$filter"] == "Enabled eq true"
    assert params["$select"] == "Id,Number"
    assert params["$expand"] == "Groups"
    assert params["$orderby"] == "Number"
    assert params["$top"] == 10
    assert params["$skip"] == 20
    assert params["$count"] == "true"


def test_to_params_returns_copy():
    q = ODataQuery().top(5)
    p1 = q.to_params()
    p2 = q.to_params()
    p1["$top"] = 999
    assert p2["$top"] == 5


def test_repr():
    r = repr(ODataQuery().top(5))
    assert "$top" in r



from __future__ import annotations

from typing import Any, Dict, Optional

from ..models.system import LicenseStatus, SystemParameters, SystemStatus
from ..odata import ODataQuery
from .base import BaseService


class SystemService(BaseService):
    """Read-only access to system status, license, and global PBX parameters.

    Reference paths:
      GET /SystemStatus
      GET /LicenseStatus
      GET /Parameters
      GET /Parameters/Pbx.GetSystemStatus()
    """

    def get_status(self) -> SystemStatus:
        """Return current PBX system status (CPU, memory, call counts, etc.)."""
        data = self._get("/SystemStatus")
        return SystemStatus.model_validate(data)

    def get_license(self) -> LicenseStatus:
        """Return the current license / subscription status."""
        data = self._get("/LicenseStatus")
        return LicenseStatus.model_validate(data)

    def get_parameters(self, query: Optional[ODataQuery] = None) -> SystemParameters:
        """Return global PBX parameters (FQDN, ports, IPs, etc.)."""
        data = self._get("/Parameters", params=self._query_params(query))
        return SystemParameters.model_validate(data)

    def get_raw(self, path: str, params: Optional[Dict[str, Any]] = None) -> Any:
        """Escape hatch: perform an arbitrary GET against the XAPI."""
        return self._get(path, params=params)

    def post_raw(self, path: str, json: Any = None) -> Any:
        """Escape hatch: perform an arbitrary POST against the XAPI."""
        return self._post(path, json=json)



from __future__ import annotations

from typing import Any


class ODataQuery:
    """Fluent builder for OData v4 query parameters.

    Example::

        q = (
            ODataQuery()
            .filter("FirstName eq 'John'")
            .select("Id", "FirstName", "LastName", "Email")
            .expand("Groups")
            .order_by("LastName")
            .top(25)
            .skip(0)
            .count()
        )
        users = client.users.list(query=q)
    """

    def __init__(self) -> None:
        self._params: dict[str, Any] = {}

    # --- pagination -----------------------------------------------------------

    def top(self, n: int) -> ODataQuery:
        self._params["$top"] = n
        return self

    def skip(self, n: int) -> ODataQuery:
        self._params["$skip"] = n
        return self

    def count(self, include: bool = True) -> ODataQuery:
        self._params["$count"] = str(include).lower()
        return self

    # --- projection / expansion -----------------------------------------------

    def select(self, *fields: str) -> ODataQuery:
        self._params["$select"] = ",".join(fields)
        return self

    def expand(self, *relations: str) -> ODataQuery:
        self._params["$expand"] = ",".join(relations)
        return self

    # --- filtering / ordering -------------------------------------------------

    def filter(self, expression: str) -> ODataQuery:
        self._params["$filter"] = expression
        return self

    def search(self, phrase: str) -> ODataQuery:
        self._params["$search"] = phrase
        return self

    def order_by(self, *fields: str) -> ODataQuery:
        self._params["$orderby"] = ",".join(fields)
        return self

    # --- serialisation --------------------------------------------------------

    def to_params(self) -> dict[str, Any]:
        return dict(self._params)

    def __repr__(self) -> str:
        return f"ODataQuery({self._params!r})"



from __future__ import annotations

from typing import Any, Dict, Iterator, List, Optional

from ..models.queues import RingGroup, RingGroupMember
from ..odata import ODataQuery
from .base import BaseService


class RingGroupsService(BaseService):
    """CRUD operations for ring groups.

    Reference paths:
      GET    /RingGroups
      POST   /RingGroups
      GET    /RingGroups({Id})
      PATCH  /RingGroups({Id})
      DELETE /RingGroups({Id})
    """

    _PATH = "/RingGroups"

    def list(self, query: Optional[ODataQuery] = None) -> List[RingGroup]:
        data = self._list_raw(self._PATH, query)
        return [RingGroup.model_validate(item) for item in data.get("value", [])]

    def iterate(self, query: Optional[ODataQuery] = None) -> Iterator[RingGroup]:
        yield from self._paginate(self._PATH, RingGroup, query)

    def create(self, ring_group: RingGroup | Dict[str, Any]) -> RingGroup:
        payload = (
            ring_group.model_dump(by_alias=True, exclude_none=True)
            if isinstance(ring_group, RingGroup)
            else ring_group
        )
        data = self._post(self._PATH, json=payload)
        return RingGroup.model_validate(data)

    def get(self, group_id: int, query: Optional[ODataQuery] = None) -> RingGroup:
        data = self._get(f"{self._PATH}({group_id})", params=self._query_params(query))
        return RingGroup.model_validate(data)

    def update(self, group_id: int, changes: RingGroup | Dict[str, Any]) -> None:
        payload = (
            changes.model_dump(by_alias=True, exclude_none=True)
            if isinstance(changes, RingGroup)
            else changes
        )
        self._patch(f"{self._PATH}({group_id})", json=payload)

    def delete(self, group_id: int, etag: Optional[str] = None) -> None:
        self._delete(f"{self._PATH}({group_id})", etag=etag)



from __future__ import annotations

from typing import Any, Generic, List, Optional, TypeVar

from pydantic import BaseModel, ConfigDict, Field

T = TypeVar("T")


class _Base(BaseModel):
    model_config = ConfigDict(
        populate_by_name=True,
        extra="allow",
    )


class ODataCollection(_Base, Generic[T]):
    """Generic OData collection response envelope."""

    odata_context: Optional[str] = Field(None, alias="@odata.context")
    odata_count: Optional[int] = Field(None, alias="@odata.count")
    odata_next_link: Optional[str] = Field(None, alias="@odata.nextLink")
    value: List[T] = Field(default_factory=list)

    @property
    def count(self) -> Optional[int]:
        return self.odata_count

    @property
    def next_link(self) -> Optional[str]:
        return self.odata_next_link


class ODataError(_Base):
    code: Optional[str] = None
    message: Optional[str] = None
    target: Optional[str] = None
    details: Optional[List[Any]] = None
    innererror: Optional[Any] = None



from __future__ import annotations

from datetime import datetime
from typing import Any, Dict, Iterator, List, Optional

from ..odata import ODataQuery
from .base import BaseService


def _fmt_dt(dt: datetime) -> str:
    """Format a datetime to the ISO-8601 pattern expected by 3CX path parameters."""
    return dt.strftime("%Y-%m-%dT%H:%M:%S") + "+00:00"


class ReportsService(BaseService):
    """Access 3CX reporting endpoints.

    All report methods return raw dicts — the raw API JSON is returned because
    the report schemas are highly specialised.  Pass the result to a Pydantic
    model of your choice or consume the dict directly.

    Reference paths (selection):
      GET /ReportCallLogData/Pbx.GetCallLogData(...)
      GET /ReportExtensionStatistics/Pbx.GetExtensionStatistics(...)
      GET /ReportQueuePerformanceOverview/Pbx.GetQueuePerformanceOverview(...)
      GET /ReportAgentLoginHistory/Pbx.GetAgentLoginHistory(...)
      GET /ActivityLog/Pbx.GetLogs(...)
      GET /ActivityLog/Pbx.GetFilter()
    """

    # ------------------------------------------------------------------
    # Call log
    # ------------------------------------------------------------------

    def get_call_log(
        self,
        start: datetime,
        end: datetime,
        *,
        query: Optional[ODataQuery] = None,
    ) -> List[Dict[str, Any]]:
        """Return call log entries between *start* and *end* (UTC)."""
        path = (
            f"/ReportCallLogData/Pbx.GetCallLogData("
            f"periodFrom={_fmt_dt(start)},periodTo={_fmt_dt(end)})"
        )
        data = self._get(path, params=self._query_params(query))
        return data.get("value", data)

    # ------------------------------------------------------------------
    # Extension statistics
    # ------------------------------------------------------------------

    def get_extension_statistics(
        self,
        start: datetime,
        end: datetime,
        *,
        query: Optional[ODataQuery] = None,
    ) -> List[Dict[str, Any]]:
        path = (
            f"/ReportExtensionStatistics/Pbx.GetExtensionStatistics("
            f"periodFrom={_fmt_dt(start)},periodTo={_fmt_dt(end)})"
        )
        data = self._get(path, params=self._query_params(query))
        return data.get("value", data)

    # ------------------------------------------------------------------
    # Queue performance
    # ------------------------------------------------------------------

    def get_queue_performance_overview(
        self,
        start: datetime,
        end: datetime,
        *,
        query: Optional[ODataQuery] = None,
    ) -> List[Dict[str, Any]]:
        path = (
            f"/ReportQueuePerformanceOverview/Pbx.GetQueuePerformanceOverview("
            f"periodFrom={_fmt_dt(start)},periodTo={_fmt_dt(end)})"
        )
        data = self._get(path, params=self._query_params(query))
        return data.get("value", data)

    def get_queue_performance_totals(
        self,
        start: datetime,
        end: datetime,
        *,
        query: Optional[ODataQuery] = None,
    ) -> List[Dict[str, Any]]:
        path = (
            f"/ReportQueuePerformanceTotals/Pbx.GetQueuePerformanceTotals("
            f"periodFrom={_fmt_dt(start)},periodTo={_fmt_dt(end)})"
        )
        data = self._get(path, params=self._query_params(query))
        return data.get("value", data)

    # ------------------------------------------------------------------
    # Agent login history
    # ------------------------------------------------------------------

    def get_agent_login_history(
        self,
        start: datetime,
        end: datetime,
        *,
        query: Optional[ODataQuery] = None,
    ) -> List[Dict[str, Any]]:
        path = (
            f"/ReportAgentLoginHistory/Pbx.GetAgentLoginHistory("
            f"periodFrom={_fmt_dt(start)},periodTo={_fmt_dt(end)})"
        )
        data = self._get(path, params=self._query_params(query))
        return data.get("value", data)

    # ------------------------------------------------------------------
    # Activity log
    # ------------------------------------------------------------------

    def get_activity_log_filter(self) -> Dict[str, Any]:
        return self._get("/ActivityLog/Pbx.GetFilter()")

    def get_activity_logs(
        self,
        start: datetime,
        end: datetime,
        *,
        extension: Optional[str] = None,
        call: Optional[str] = None,
        severity: Optional[str] = None,
        query: Optional[ODataQuery] = None,
    ) -> List[Dict[str, Any]]:
        ext = f"'{extension}'" if extension else "null"
        cal = f"'{call}'" if call else "null"
        sev = f"'{severity}'" if severity else "null"
        path = (
            f"/ActivityLog/Pbx.GetLogs("
            f"startDate={_fmt_dt(start)},endDate={_fmt_dt(end)},"
            f"extension={ext},call={cal},severity={sev})"
        )
        data = self._get(path, params=self._query_params(query))
        return data.get("value", data)

    def purge_activity_logs(self) -> None:
        self._post("/ActivityLog/Pbx.PurgeLogs")

    # ------------------------------------------------------------------
    # Abandoned queue calls
    # ------------------------------------------------------------------

    def get_abandoned_queue_calls(
        self,
        start: datetime,
        end: datetime,
        *,
        query: Optional[ODataQuery] = None,
    ) -> List[Dict[str, Any]]:
        path = (
            f"/ReportAbandonedQueueCalls/Pbx.GetAbandonedQueueCalls("
            f"periodFrom={_fmt_dt(start)},periodTo={_fmt_dt(end)})"
        )
        data = self._get(path, params=self._query_params(query))
        return data.get("value", data)



"""Integration-style tests: verify that HTTP error codes become the right exceptions."""
from __future__ import annotations

import pytest
from pytest_httpx import HTTPXMock

from tests.conftest import TOKEN_URL, TOKEN_JSON, api
from threecx import ThreeCXClient
from threecx.exceptions import (
    AuthenticationError,
    NotFoundError,
    RateLimitError,
    ServerError,
    ValidationError,
)

ERROR_BODY = {"error": {"message": "Something went wrong", "code": "GeneralException"}}


@pytest.fixture
def client(httpx_mock: HTTPXMock) -> ThreeCXClient:
    httpx_mock.add_response(url=TOKEN_URL, method="POST", json=TOKEN_JSON, is_reusable=True)
    return ThreeCXClient("https://pbx.test", "id", "secret")


@pytest.mark.parametrize("status_code,exc_type", [
    (400, ValidationError),
    (401, AuthenticationError),
    (404, NotFoundError),
    (422, ValidationError),
    (429, RateLimitError),
    (500, ServerError),
    (503, ServerError),
])
def test_error_status_raises_correct_exception(
    client: ThreeCXClient,
    httpx_mock: HTTPXMock,
    status_code: int,
    exc_type: type,
) -> None:
    httpx_mock.add_response(url=api("/Users"), status_code=status_code, json=ERROR_BODY)
    with pytest.raises(exc_type) as exc_info:
        client.users.list()
    assert exc_info.value.status_code == status_code


def test_not_found_on_get_user(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=api("/Users(9999)"),
        status_code=404,
        json={"error": {"message": "User 9999 not found"}},
    )
    with pytest.raises(NotFoundError, match="User 9999 not found"):
        client.users.get(9999)


def test_validation_error_on_create(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=api("/Users"),
        method="POST",
        status_code=400,
        json={"error": {"message": "Number is already in use"}},
    )
    with pytest.raises(ValidationError, match="Number is already in use"):
        client.users.create({"Number": "100"})


def test_server_error_carries_status_code(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(url=api("/SystemStatus"), status_code=503, json=ERROR_BODY)
    with pytest.raises(ServerError) as exc_info:
        client.system.get_status()
    assert exc_info.value.status_code == 503


def test_plain_text_error_body(client: ThreeCXClient, httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        url=api("/Users"),
        status_code=500,
        text="Internal Server Error",
    )
    with pytest.raises(ServerError):
        client.users.list()



"""Unit tests for Pydantic model validation, alias mapping, and serialisation."""
from __future__ import annotations

import pytest

from threecx.models.calls import ActiveCall, CallHistoryEntry
from threecx.models.contacts import Contact
from threecx.models.phones import Phone, PhoneTemplate
from threecx.models.queues import Queue, QueueAgent, RingGroup, RingGroupMember
from threecx.models.system import LicenseStatus, SystemStatus
from threecx.models.trunks import Peer, Sbc, Trunk
from threecx.models.users import ForwardingProfile, Greeting, Group, User


# --- User ---------------------------------------------------------------------

def test_user_validates_from_api_payload():
    payload = {
        "Id": 101,
        "Number": "100",
        "FirstName": "Alice",
        "LastName": "Smith",
        "Email": "alice@example.com",
        "Enabled": True,
    }
    user = User.model_validate(payload)
    assert user.id == 101
    assert user.number == "100"
    assert user.first_name == "Alice"
    assert user.last_name == "Smith"
    assert user.email == "alice@example.com"
    assert user.enabled is True


def test_user_full_name():
    user = User.model_validate({"FirstName": "Bob", "LastName": "Jones"})
    assert user.full_name == "Bob Jones"


def test_user_full_name_only_first():
    user = User.model_validate({"FirstName": "Bob"})
    assert user.full_name == "Bob"


def test_user_serialises_with_aliases():
    user = User(Number="200", FirstName="Carol", Email="carol@x.com")
    payload = user.model_dump(by_alias=True, exclude_none=True)
    assert payload["Number"] == "200"
    assert payload["FirstName"] == "Carol"
    assert "Id" not in payload  # excluded because None


def test_user_extra_fields_allowed():
    user = User.model_validate({"Id": 1, "Number": "100", "UnknownField": "ignored"})
    assert user.id == 1


def test_user_nested_groups():
    payload = {
        "Id": 1,
        "Number": "100",
        "Groups": [{"Id": 5, "Name": "Sales"}],
    }
    user = User.model_validate(payload)
    assert len(user.groups) == 1
    assert user.groups[0].name == "Sales"


# --- Group / ForwardingProfile / Greeting ------------------------------------

def test_group_validates():
    g = Group.model_validate({"Id": 3, "Name": "Support"})
    assert g.id == 3
    assert g.name == "Support"


def test_forwarding_profile_validates():
    fp = ForwardingProfile.model_validate({"Id": 1, "Name": "Away", "Enabled": False})
    assert fp.enabled is False


def test_greeting_validates():
    gr = Greeting.model_validate({"Id": 7, "GreetingFile": "greeting.wav", "GreetingType": "Voicemail"})
    assert gr.filename == "greeting.wav"


# --- ActiveCall ---------------------------------------------------------------

def test_active_call_validates():
    payload = {
        "Id": 42,
        "Status": "Connected",
        "Established": True,
        "Caller": "100",
        "Callee": "200",
        "Duration": 65,
        "OnHold": False,
    }
    call = ActiveCall.model_validate(payload)
    assert call.id == 42
    assert call.caller == "100"
    assert call.duration == 65
    assert call.on_hold is False


def test_active_call_optional_fields_default_none():
    call = ActiveCall.model_validate({"Id": 1})
    assert call.recording is None
    assert call.is_internal is None


# --- CallHistoryEntry ---------------------------------------------------------

def test_call_history_entry_validates():
    payload = {
        "Id": "abc-123",
        "Caller": "100",
        "Callee": "0031612345678",
        "CallDirection": "Outbound",
        "Status": "Answered",
        "TalkDuration": 120,
    }
    entry = CallHistoryEntry.model_validate(payload)
    assert entry.id == "abc-123"
    assert entry.direction == "Outbound"
    assert entry.talk_duration == 120


# --- Queue --------------------------------------------------------------------

def test_queue_validates():
    payload = {
        "Id": 10,
        "Number": "800",
        "Name": "Support Queue",
        "RingTimeout": 30,
        "SLATime": 20,
    }
    q = Queue.model_validate(payload)
    assert q.name == "Support Queue"
    assert q.ring_timeout == 30


def test_queue_agent_validates():
    agent = QueueAgent.model_validate({"DnNumber": "101", "Position": 1, "WrapUpTime": 10})
    assert agent.dn_number == "101"
    assert agent.wrap_up_time == 10


# --- RingGroup ----------------------------------------------------------------

def test_ring_group_validates():
    payload = {
        "Id": 20,
        "Number": "700",
        "Name": "Sales Ring",
        "RingStrategy": "Hunt",
        "RingTime": 15,
    }
    rg = RingGroup.model_validate(payload)
    assert rg.ring_strategy == "Hunt"


def test_ring_group_member_validates():
    m = RingGroupMember.model_validate({"DnNumber": "102", "Timeout": 20})
    assert m.dn_number == "102"


# --- Trunk / Peer / Sbc -------------------------------------------------------

def test_trunk_validates():
    trunk = Trunk.model_validate({"Id": 1, "Name": "SIP Trunk A", "Host": "sip.carrier.com", "Port": 5060})
    assert trunk.host == "sip.carrier.com"


def test_peer_validates():
    peer = Peer.model_validate({"Id": 2, "Name": "Gateway", "Host": "192.168.1.1"})
    assert peer.host == "192.168.1.1"


def test_sbc_validates():
    sbc = Sbc.model_validate({"Id": 3, "Name": "SBC-1", "Host": "10.0.0.1"})
    assert sbc.name == "SBC-1"


# --- Contact ------------------------------------------------------------------

def test_contact_validates():
    payload = {
        "Id": 99,
        "FirstName": "Dave",
        "LastName": "Brown",
        "Email": "dave@example.com",
        "PhoneNumber": "+31612345678",
        "ContactType": "3CX",
    }
    c = Contact.model_validate(payload)
    assert c.full_name == "Dave Brown"
    assert c.phone_number == "+31612345678"


# --- Phone / PhoneTemplate ----------------------------------------------------

def test_phone_validates():
    phone = Phone.model_validate({
        "Id": 5,
        "MAC": "AA:BB:CC:DD:EE:FF",
        "Model": "Yealink T48S",
        "IPAddress": "192.168.1.50",
    })
    assert phone.mac == "AA:BB:CC:DD:EE:FF"
    assert phone.model == "Yealink T48S"


def test_phone_template_validates():
    tmpl = PhoneTemplate.model_validate({"Id": 1, "Name": "T48S Default", "Manufacturer": "Yealink"})
    assert tmpl.manufacturer == "Yealink"


# --- SystemStatus / LicenseStatus ---------------------------------------------

def test_system_status_validates():
    payload = {
        "HasActiveCalls": True,
        "CallsActive": 3,
        "ExtensionsRegistered": 25,
        "ExtensionsTotal": 30,
        "CpuUsage": 12.5,
        "Version": "20.0.0.123",
    }
    s = SystemStatus.model_validate(payload)
    assert s.calls_active == 3
    assert s.pbx_version == "20.0.0.123"


def test_license_status_validates():
    lic = LicenseStatus.model_validate({
        "ProductCode": "3CXPSPROF",
        "MaxSimCalls": 32,
        "MaxExtensions": 100,
    })
    assert lic.max_simultaneous_calls == 32
